/// Josh Unit Test Library

#pragma once
#include <functional>
#include <string>
#include <vector>
#include <jlog/jlog.hpp>

//#define NDEBUG for disabling assert, but tests still end up passing.

// TODO: Move implementation to jtest::detail
// have this file primarily expose the macros intended for users

namespace jtest {

    /// Structure to store test meta-data, as tests are initially registered, and ran later.
    struct testdef
    {
        std::string testname;
        std::function<void()> callback;
        std::string file;
        int line;
        bool passed;
    };

    /// jlog log wrapper for jtest
    void log(std::vector<jlog::token> tokens);

    /// Generates and returns a sequence of logger tokens pre-formatted to the test data.
    std::vector<jlog::token> log_test_format(const std::string& testname, const std::string& file, int line, bool passed);

    /// Generates and returns a sequence of logger tokens for the test summary printed at the end of testing.
    std::vector<jlog::token> log_test_tracking_format();

    /// Registers a test internally.
    void definetest(const std::string& testname, const std::function<void()>& callback, const std::string& file, int line);


    // TODO: implement streaming a custom failure message with << operator on check statements

    /// Raises an exception if the given condition is false, otherwise returns true.
    bool check(bool condition);

    /// Raises an exception of the given values evaluate to not-equal, otherwise returns true.
    template <typename T>
    bool check_eq(T a, T b) {
        if (a != b)
            throw std::runtime_error("Test check failed!!");
        return true;
    }

    /// Raises an exception if the given floats are not equal, up to the given epsilon. Otherwise returns true.
    /// @param epsilon The accuracy required to pass the test.
    bool check_float_eq(float a, float b, float epsilon = 1e-3f);

    /// Raises an exception if the given doubles are not equal, up to the given epsilon. Otherwise returns true.
    /// @param epsilon The accuracy required to pass the test.
    bool check_double_eq(double a, double b, double epsilon = 1e-3f);

    /// Raises an exception if the given strings are not equal, otherwise returns true.
    bool check_string_eq(const std::string& a, const std::string& b);

    /// Runs a given test, generates a report, and returns the test result as a boolean.
    bool test(const std::string& testname, const std::function<void()>& callback, const std::string& file, int line);

    /// Runs all tests that have been registered, and generates a final summary report for the testing suite.
    void run_tests();
}

/// TEST macro to be used by API consumers. Automatically grabs reflection data such as current file, line, function name, etc.
#define TEST(a, b) jtest::definetest(a, b, __FILE__, __LINE__);

/// TODO: Implement check macros
